/*
 * vim: ts=4 sw=4 fdm=marker noet tw=78 ai
 * Copyright (c) 2012, Mel Flynn
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/cdefs.h>
#ifdef __FBSDID
__FBSDID("$Coar: trunk/ldapenter/attributes.c 2 2012-01-10 23:05:07Z rflynn@acsalaska.net $");
#endif
#include <sys/queue.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>

#include <ldap.h>
#include <ldap_schema.h>

#include "extern.h"
#include "search.h"
#include "attributes.h"

static int lde_add_attribute_name(runtime_t *, char *);
static attr_entry_t *lde_attribute_new(char *, char *, size_t);

/*
 * Add an attribute value
 * To support multi valued attributes, we do add duplicates. However, if the
 * name is already available in the attribute list /and/ it's value is NULL,
 * then we set that attribute's value.
 * NOTE:
 * Caller is responsible for initialization of the attribute list, as it
 * allows caller to also clear it when needed.
 */
int
lde_add_attribute_value(runtime_t *ctx, char *name, struct berval *value)
{
	int retval = LDAP_SUCCESS;
	attr_entry_t *new;

	assert(ctx != NULL);
	assert(ctx->atl != NULL);
	TRACEIN(ctx, __func__, "Adding value for %s: %s", name, value->bv_val);

	new = lde_find_attribute(ctx, name);
	if( new == NULL || new->value != NULL )
	{
		LOG(ctx, 2, "Not found or empty value");
		new = lde_attribute_new(name, value->bv_val, (size_t)value->bv_len);
		if( new == NULL )
		{
			retval = LDAP_OPERATIONS_ERROR;
			goto out;
		}
		LOG(ctx, 2, "Inserting %s: %s (%zu)", new->name, new->value,
				new->value_len);
		STAILQ_INSERT_TAIL(ctx->atl, new, next);
	}
	else
	{
		if( (new->value = malloc(value->bv_len)) == NULL )
		{
			retval = LDAP_OPERATIONS_ERROR;
			goto out;
		}
		(void)memcpy(new->value, value->bv_val, value->bv_len);
		new->value_len = value->bv_len;
	}

out:
	TRACEOUT(ctx, __func__, "Returning 0x%02x", retval);
	return (retval);
}

attr_entry_t *
lde_find_attribute(runtime_t *ctx, char *name)
{
	attr_entry_t *cur, *retval = NULL;

	assert(ctx != NULL);
	assert(name != NULL);
	TRACEIN(ctx, __func__, "Looking for %s", name);

	if( ctx->atl == NULL )
		goto out;

	STAILQ_FOREACH(cur, ctx->atl, next)
	{
		if( strcasecmp(cur->name, name) == 0 )
		{
			retval = cur;
			break;
		}
	}

out:
	TRACEOUT(ctx, __func__, "Returning %p", retval);
	return(retval);
}

int
lde_fetch_class_attributes(runtime_t *ctx, char *oc_name)
{
	int retval; // Return value
	int i; // iterator
	class_entry_t *cur; // current object class

	assert(ctx != NULL);
	assert(oc_name != NULL);
	TRACEIN(ctx, __func__, "Looking for class: %s", oc_name);

	SLIST_FOREACH(cur, ctx->ocl, next)
	{
		if( strcasecmp(cur->name, oc_name) == 0 )
		{
			/*
			 * We skip abstract classes:
			 * - They should be implemented, so we will always encounter one
			 *   that does the implementation.
			 * - And as such we might as well skip the duplicate
			 */
			if( cur->lo->oc_kind == LDAP_SCHEMA_ABSTRACT )
				break;

			/* Required attributes */
			if( cur->lo->oc_at_oids_must != NULL )
			{
				for( i=0; cur->lo->oc_at_oids_must[i] != NULL; i++ )
				{
					retval = lde_add_attribute_name(ctx,
							cur->lo->oc_at_oids_must[i]);
					if( retval != LDAP_SUCCESS )
						goto out;
				}
			}
			/* Optional attributes */
			if( cur->lo->oc_at_oids_may != NULL )
			{
				for( i=0; cur->lo->oc_at_oids_may[i] != NULL; i++ )
				{
					retval = lde_add_attribute_name(ctx,
							cur->lo->oc_at_oids_may[i]);
					if( retval != LDAP_SUCCESS )
						goto out;
				}
			}
			/* Recurse superior objects */
			if( cur->lo->oc_sup_oids != NULL )
			{
				for( i=0; cur->lo->oc_sup_oids[i] != NULL; i++ )
				{
					retval = lde_fetch_class_attributes(ctx,
							cur->lo->oc_sup_oids[i]);
					if( retval != LDAP_SUCCESS )
						goto out;
				}
			}
		}
	}

out:
	TRACEOUT(ctx, __func__, "Returning %02x", retval);
	return (retval);
}

/*
 * Add an attribute's name.
 * We do not add:
 * - Duplicates
 * - the "objectClass" attribute
 */
static int
lde_add_attribute_name(runtime_t *ctx, char *attr_name)
{
	int retval = LDAP_SUCCESS; // Return value
	attr_entry_t *cur; // current attribute
	attr_entry_t *new; // new attribute

	assert(ctx != NULL);
	assert(attr_name != NULL);
	assert(ctx->atl != NULL);
	TRACEIN(ctx, __func__, "Adding attribute %s", attr_name);

	// Do not add the "objectClass" as an attribute
	if( ATTR_IS_OC(attr_name) )
		goto out;

	STAILQ_FOREACH(cur, ctx->atl, next)
	{
		if( strcasecmp(cur->name, attr_name) == 0 )
			goto out;
	}
	if( (new = malloc(sizeof(attr_entry_t))) == NULL )
	{
		retval = LDAP_OPERATIONS_ERROR;
		goto out;
	}
	memset(new, 0, sizeof(attr_entry_t));

	new->name = strdup(attr_name);
	new->name_len = strlen(attr_name);

	LOG(ctx, 2, "Adding attribute %s", new->name);
	ctx->at_count++;
	STAILQ_INSERT_TAIL(ctx->atl, new, next);

out:
	TRACEOUT(ctx, __func__, "Returning %02x", retval);
	return (retval);
}

static attr_entry_t *
lde_attribute_new(char *name, char *value, size_t val_len)
{
	attr_entry_t *retval;

	assert(name != NULL);
	if( (retval = malloc(sizeof(attr_entry_t))) != NULL )
	{
		memset(retval, 0, sizeof(attr_entry_t));
		retval->name = strdup(name);
		retval->name_len = strlen(name);
		if( val_len )
		{
			if( (retval->value = malloc(val_len+1)) == NULL )
			{
				free(retval->name);
				free(retval);
				retval = NULL;
			}
			else
			{
				(void)memcpy(retval->value, value, val_len);
				retval->value[val_len] = '\0';
				retval->value_len = val_len;
			}
		}
	}

	return(retval);
}

void
lde_attribute_list_init(runtime_t *ctx)
{
	if( ctx->atl != NULL )
	{
		attr_entry_t *n1, *n2;

		n1 = STAILQ_FIRST(ctx->atl);
		while( n1 != NULL )
		{
			n2 = STAILQ_NEXT(n1, next);
			free(n1->name);
			free(n1->value);
			free(n1);
			n1 =  n2;
		}
		free(ctx->atl);
		// Not needed: ctx->atl = NULL;
	}
	ctx->atl = malloc(sizeof(attr_list_t));
	STAILQ_INIT(ctx->atl);
}
