/*
 * vim: ts=4 sw=4 fdm=marker noet tw=78 ai
 * Copyright (c) 2012, Mel Flynn
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *     Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/cdefs.h>
#ifdef __FBSDID
__FBSDID("$Coar: trunk/ldapenter/classes.c 2 2012-01-10 23:05:07Z rflynn@acsalaska.net $");
#endif
#include <sys/queue.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <err.h>

#include <ldap.h>
#include <ldap_schema.h>

#include "extern.h"
#include "search.h"
#include "classes.h"

#define UNIMPLEMENTED LDAP_SUCCESS

static const int relaxed = LDAP_SCHEMA_ALLOW_ALL;
static const int hyper   = LDAP_SCHEMA_ALLOW_NONE;

static class_entry_t * class_entry_new(LDAPObjectClass *);

int
lde_add_class(runtime_t *ctx, char *schema)
{
	LDAPObjectClass *oc;
	const char *errpos;
	class_entry_t *new, *tmp;
	int retval = LDAP_SUCCESS;

	assert(ctx != NULL);
	assert(ctx->ld != NULL);
	assert(schema != NULL);
	TRACEIN(ctx, __func__, "schemalen = %zu", strlen(schema));

	if( ctx->ocl == NULL )
	{
		if( (ctx->ocl = malloc(sizeof(class_list_t))) == NULL )
		{
			retval = LDAP_OPERATIONS_ERROR;
			goto out;
		}
		SLIST_INIT(ctx->ocl);
	}

	oc = ldap_str2objectclass(schema, &retval, &errpos, relaxed);
	if( retval != 0 )
	{
		warnx("%s", ldap_scherr2str(retval));
		warnx("At %20s", errpos);
		retval = LDAP_OPERATIONS_ERROR;
	}
	else
	{
		int found = 0;
		new = class_entry_new(oc);
		SLIST_FOREACH(tmp, ctx->ocl, next)
		{
			if( strcmp(tmp->name, new->name) == 0 )
			{
				found++;
				break;
			}
		}
		if( ! found )
		{
			LOG(ctx, 2, "Adding Objectlass: %s", new->name);
			ctx->oc_count++;
			SLIST_INSERT_HEAD(ctx->ocl, new, next);
		}
	}

out:
	TRACEOUT(ctx, __func__, "Returning 0x%02x", retval);

	return(retval);
}

static class_entry_t *
class_entry_new(LDAPObjectClass *oc)
{
	class_entry_t *retval;

	if( (retval = malloc(sizeof(class_entry_t))) == NULL )
		return NULL;

	retval->lo = oc;
	if( oc->oc_names[0] == NULL )
	{
		warnx("No name known for object class. Using OID");
		retval->name = strdup(oc->oc_oid);
	}
	else
		retval->name = strdup(oc->oc_names[0]);

	return (retval);
}
